using UnityEngine;

namespace Damage
{
    public class UltraStrengthOrganSyringe : SyringeBehaviour
    {
        public override string GetLiquidID() => UltraStrengthOrganSerumLiquid.ID;

        public class UltraStrengthOrganSerumLiquid : TemporaryBodyLiquid
        {
            public const string ID = "ULTRA STRENGTH SERUM [Advanced Damage]";

            public UltraStrengthOrganSerumLiquid()
            {
                this.Color = new Color(0.7604228f, 0.7137255f, 1f);
            }

            public override void OnEnterContainer(BloodContainer container)
            {
            }

            public override void OnEnterLimb(LimbBehaviour limb)
            {
                if (limb.SpeciesIdentity == "Android")
                    return;
                limb.Vitality *= 0.1f;
                limb.RegenerationSpeed += 5f;
                limb.ImpactPainMultiplier *= 0.1f;
                limb.InitialHealth += 150f;
                limb.Health = limb.InitialHealth;
                limb.BreakingThreshold += 10f;
                limb.BaseStrength = Mathf.Min(15f, limb.BaseStrength + 5f);
                limb.ShotDamageMultiplier *= 0.1f;
                limb.CirculationBehaviour.HealBleeding();
                if (limb.gameObject.TryGetComponent(out HumanDamageBehaviour damage))
                {
                    damage.strength = limb.BaseStrength;
                    damage.tissueDamageThreshold += 150f;
                    damage.tissueDamage = 0;
                    damage.initialBoneHealth += 150f;
                    damage.boneHealth = damage.initialBoneHealth;
                    damage.fractureChance *= 0.9f;
                    damage.initialBoneStrength *= 1.1f;
                    damage.boneStrength = damage.initialBoneStrength;
                    if (damage.hasOrgan)
                    {
                        foreach (Organ organ in damage.organs)
                        {
                            organ.initialHealth += 25f;
                            organ.health = organ.initialHealth;
                            organ.initialIntegrity += 25f;
                            organ.integrity = organ.initialIntegrity;
                            organ.damageMultiplier *= 0.95f;
                            organ.regenerationSpeed += 1f;
                        }
                    }
                }
                limb.Wince(10f);
            }

            public override void OnExitContainer(BloodContainer container)
            {
            }
        }
    }
    
    public class ZombieOrganSyringe : SyringeBehaviour
    {
        public override string GetLiquidID() => ZombiePoisonOrganLiquid.ID;

        public class ZombiePoisonOrganLiquid : Liquid
        {
            public const string ID = "REANIMATION AGENT [Advanced Damage]";

            public ZombiePoisonOrganLiquid()
            {
                this.Color = new Color(0.8835177f, 1f, 0.7122642f);
            }

            public override void OnEnterContainer(BloodContainer container)
            {
            }

            public override void OnEnterLimb(LimbBehaviour limb)
            {
                if (limb.SpeciesIdentity == "Android")
                    return;
                limb.IsZombie = true;
                limb.Health = limb.InitialHealth;
                limb.Numbness = 0.0f;
                CirculationBehaviour circulationBehaviour = limb.CirculationBehaviour;
                float amount = circulationBehaviour.GetAmount(limb.GetOriginalBloodType());
                if (amount < 0.100000001490116)
                {
                    double num = circulationBehaviour.AddLiquid(limb.GetOriginalBloodType(), 1f - amount);
                }
                circulationBehaviour.BloodFlow = 1f;
                circulationBehaviour.HealBleeding();
                circulationBehaviour.BleedingRate = 0.0f;
                circulationBehaviour.IsPump = circulationBehaviour.WasInitiallyPumping;
                if (limb.gameObject.TryGetComponent(out HumanDamageBehaviour damage))
                {
                    if (!damage.CanReanimate())
                        return;
                    damage.Revive();
                    damage.vitalOrganDestroyed = false;
                    if (damage.hasOrgan)
                    {
                        foreach (Organ organ in damage.organs)
                        {
                            organ.destroyed = false;
                            organ.integrity = Mathf.Min(organ.initialIntegrity * 0.5f, organ.integrity);
                            damage.Revive();
                            if (organ.name == "Brain")
                            {
                                organ.initialHealth = Mathf.Max(organ.initialHealth, 40f);
                                organ.health = organ.initialHealth;
                                organ.initialIntegrity = Mathf.Max(organ.initialIntegrity, 40f);
                                organ.chance *= Random.Range(0.35f, 0.9f);
                                organ.integrity = organ.initialIntegrity;
                                organ.damageMultiplier = Mathf.Min(Random.Range(0.1f, 0.9f), organ.damageMultiplier);
                            }
                        }
                    }
                }
                if (limb.RoughClassification != LimbBehaviour.BodyPart.Head)
                    return;
                limb.Person.Consciousness = 1f;
                limb.Person.ShockLevel = 0.0f;
                limb.Person.PainLevel = 0.0f;
                limb.Person.OxygenLevel = 1f;
                limb.Person.AdrenalineLevel = 1f;
            }

            public override void OnExitContainer(BloodContainer container)
            {
            }
        }
    }
}